using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;
using EmissiveClothing;

namespace EmissiveClothing
{

    public class EmissiveExtraEffects : MVRScript
    {

        protected EmissiveClothingPlus _emissivePlusScript;

        protected JSONStorableBool _pulseEnabled;

        protected JSONStorableStringChooser _pulseMainType;
        protected JSONStorableString _pulseInfoJson;

        protected JSONStorableFloat _pulseMinValue;
        protected JSONStorableFloat _pulseMaxValue;
        protected JSONStorableFloat _pulseSpeed;
        protected JSONStorableFloat _pulseCurrentStrength;

        protected JSONStorableStringChooser _pulseMorphUidChooser;
        protected JSONStorableString _pulseMorphUid;
        protected JSONStorableBool _pulseMorphInverse;
        protected JSONStorableFloat _pulseMorphValueModifier;
        protected JSONStorableFloat _pulseMorphValueOffset;
        protected List<string> _filterMorphNames = new List<string>();
        protected List<string> _filterMorphUids = new List<string>();
        protected DAZMorph _cachedPulseMorph = null;
        protected float _cachedPulseMorphTime;

        protected JSONStorableStringChooser _pulseAtomUidChooser;
        protected JSONStorableString _pulseAtomUid;
        protected JSONStorableStringChooser _pulseAtomStoreableUidChooser;
        protected JSONStorableString _pulseAtomStoreableUid;
        protected JSONStorableStringChooser _pulseAtomStoreableParamValueUidChooser;
        protected JSONStorableString _pulseAtomStoreableParamValueUid;
        protected JSONStorableBool _pulseAtomStoreableParamInverse;
        protected JSONStorableFloat _pulseAtomStoreableParamValueModifier;
        protected JSONStorableFloat _pulseAtomStoreableParamValueOffset;

        protected JSONStorableBool _colorTweeningEnabled;
        protected JSONStorableFloat _colorTweenSpeed;
        protected JSONStorableStringChooser _colorTweenType;
        protected JSONStorableColor _colorTweenA, _colorTweenB, _colorTweenC;
        protected JSONStorableBool _colorTweenWrapHue;
        protected JSONStorableColor _previewCurrentTweenColor;

        protected string _pulseParam_atomID;
        protected string _pulseParam_storeableID;
        protected string _pulseParam_paramID;
        protected Atom _sourceParamAtom;
        protected JSONStorableFloat _sourceParamValue;

        protected float _pulseTime;
        protected bool _initScriptDone = false;
        protected float _tweenCurrentTime;

        protected bool _prevPulseActive = false;
        protected float? _originalColorAlphaMultiplier;
        protected bool _prevColorTweeningActive = false;
        protected bool? _originalUseOriginalColor;
        protected HSVColor? _originalColor;

        const float PI2 = (float)(Math.PI * 2.0);
        const float PI = (float)Math.PI;

        protected Blazedust.CUAM.UIHelper _pulseGUI;

        public override void Init()
        {
            _initScriptDone = false;

            //_emissivePlusScript = FindEmissiveClothingPlus();
            //if (_emissivePlusScript == null)
            //{
            //    SuperController.LogError("EmissivePulseEffect: Unable to find EmissiveClothingPlus script");
            //    JSONStorableString infoJson = new JSONStorableString("infoText", "Unable to find EmissiveClothingPlus script. This plugin will have no effect.");
            //    CreateTextField(infoJson, false);
            //}

            // Pulse type + info
            {
                Color cPulse = new Color(0.9f, 0.9f, 1.0f);

                _pulseEnabled = new JSONStorableBool("PulseEnabled", false, (bool _) => SettingsUpdated());
                RegisterBool(_pulseEnabled);
                var toggle = CreateToggle(_pulseEnabled, true);
                toggle.label = "Pulse Enabled";
                toggle.height = 80;
                toggle.backgroundColor = cPulse;

                _pulseInfoJson = new JSONStorableString("pulseInfoText", "");
                CreateTextField(_pulseInfoJson, true);

                _pulseGUI = new Blazedust.CUAM.UIHelper(this);

                List<string> pulseMainTypes = new List<string>();
                pulseMainTypes.Add("Cyclic");
                pulseMainTypes.Add("ByMorphValue");
                pulseMainTypes.Add("ByFloatParamSource");
                _pulseMainType = new JSONStorableStringChooser("PulseMainType", pulseMainTypes, "Cyclic", "PulseMainType", (string type) => {
                    if (_initScriptDone)
                        BuildPulseGUI();
                    SettingsUpdated();
                });
                RegisterStringChooser(_pulseMainType);
                var popup = CreateScrollablePopup(_pulseMainType, true);
                popup.label = "Pulse Type";
            }

            // Cyclic
            {
                _pulseMinValue = new JSONStorableFloat("PulseMinValue", 0f, (float _) =>
                {
                    if (!_initScriptDone)
                        return;

                    if (_pulseMaxValue.val < _pulseMinValue.val)
                        _pulseMaxValue.val = _pulseMinValue.val;

                    SettingsUpdated();
                }, 0f, 1f, false, true);
                RegisterFloat(_pulseMinValue);

                _pulseMaxValue = new JSONStorableFloat("PulseMaxValue", 1f, (float _) =>
                {
                    if (!_initScriptDone)
                        return;

                    if (_pulseMinValue.val > _pulseMaxValue.val)
                        _pulseMinValue.val = _pulseMaxValue.val;

                    SettingsUpdated();
                }, 0f, 1f, false, true);
                RegisterFloat(_pulseMaxValue);

                _pulseSpeed = new JSONStorableFloat("PulseSpeed", 0.5f, (float _) => {
                    if (_pulseSpeed.val < 0f)
                        _pulseSpeed.valNoCallback = 0f;
                    SettingsUpdated();
                }, 0f, 10f, true, true);
                RegisterFloat(_pulseSpeed);

                _pulseCurrentStrength = new JSONStorableFloat("PulseCurrent", 1f, (float _) => { }, 0f, 1f, true, false);
            }

            // morph source
            {

                _pulseMorphUid = new JSONStorableString("PulseByMorphUid", "");
                RegisterString(_pulseMorphUid);

                _pulseMorphUidChooser = new JSONStorableStringChooser("PulseByMorphUidChooser", RebuildActiveMorphNames(), "", "Pulse Morph", (string morph) => {
                    if (!_initScriptDone)
                        return;

                    int index = _filterMorphNames.IndexOf(morph);
                    if (index >= 0)
                    {
                        SuperController.LogMessage("Selected morph uid: " + _filterMorphUids[index]);
                        _pulseMorphUid.val = _filterMorphUids[index];
                    }
                    else
                    {
                        _pulseMorphUid.val = "";
                    }
                    SettingsUpdated();
                });
                RegisterStringChooser(_pulseMorphUidChooser);

                _pulseMorphInverse = new JSONStorableBool("PulseByMorphInverse", false, (bool _) => SettingsUpdated());
                RegisterBool(_pulseMorphInverse);
                
                _pulseMorphValueModifier = new JSONStorableFloat("PulseMorphValueModifier", 1f, (float _) => { SettingsUpdated(); }, 0f, 2f, false, true);
                RegisterFloat(_pulseMorphValueModifier);

                _pulseMorphValueOffset = new JSONStorableFloat("PulseMorphValueOffsett", 0f, (float _) => { SettingsUpdated(); }, -1f, 1f, false, true);
                RegisterFloat(_pulseMorphValueOffset);
            }

            // float param source
            {
                
                //protected JSONStorableStringChooser _pulseAtomStoreableParamUidChooser;
                //protected JSONStorableString _pulseAtomStoreableParamUid;

                _pulseAtomUid = new JSONStorableString("PulseByFloatParamAtomUid", "");
                RegisterString(_pulseAtomUid);

                _pulseAtomUidChooser = new JSONStorableStringChooser("PulseByFloatParamAtomUidChooser", GetAtomChocies(), "", "Source Atom", (string atomID) => {
                    if (!_initScriptDone)
                        return;

                    _pulseAtomUid.val = atomID;
                    SettingsUpdated();
                });
                RegisterStringChooser(_pulseAtomUidChooser);


                _pulseAtomStoreableUid = new JSONStorableString("PulseByFloatParamStoreableUid", "");
                RegisterString(_pulseAtomStoreableUid);

                _pulseAtomStoreableUidChooser = new JSONStorableStringChooser("PulseByFloatParamStoreableUidChooser", GetAtomStoreableChoices(""), "", "Source Storeable", (string storeableID) => {
                    if (!_initScriptDone)
                        return;

                    _pulseAtomStoreableUid.val = storeableID;
                    SettingsUpdated();
                });
                RegisterStringChooser(_pulseAtomStoreableUidChooser);



                _pulseAtomStoreableParamValueUid = new JSONStorableString("PulseByFloatParamValueUid", "");
                RegisterString(_pulseAtomStoreableParamValueUid);

                _pulseAtomStoreableParamValueUidChooser = new JSONStorableStringChooser("PulseByFloatParamValueUidChooser", GetAtomStoreableFloatParamNames("", ""), "", "Source Param Value", (string paramValueID) => {
                    if (!_initScriptDone)
                        return;

                    _pulseAtomStoreableParamValueUid.val = paramValueID;
                    SettingsUpdated();
                });
                RegisterStringChooser(_pulseAtomStoreableParamValueUidChooser);


                _pulseAtomStoreableParamInverse = new JSONStorableBool("PulseByFloatParamInverse", false, (bool _) => SettingsUpdated());
                RegisterBool(_pulseAtomStoreableParamInverse);

                _pulseAtomStoreableParamValueModifier = new JSONStorableFloat("PulseByFloatParamValueModifier", 1f, (float _) => { SettingsUpdated(); }, 0f, 2f, false, true);
                RegisterFloat(_pulseAtomStoreableParamValueModifier);

                _pulseAtomStoreableParamValueOffset = new JSONStorableFloat("PulseByFloatParamValueOffset", 0f, (float _) => { SettingsUpdated(); }, -1f, 1f, false, true);
                RegisterFloat(_pulseAtomStoreableParamValueOffset);
            }

            // Color tweening
            {
                Color cTween = new Color(0.9f, 1.0f, 1.0f);

                _colorTweeningEnabled = new JSONStorableBool("ColorTweeningEnabled", false, (bool _) => SettingsUpdated());
                RegisterBool(_colorTweeningEnabled);
                var toggle = CreateToggle(_colorTweeningEnabled, false);
                toggle.label = "Color Tweening Enabled";
                toggle.height = 80;
                toggle.backgroundColor = cTween;

                JSONStorableString pulseInfoJson = new JSONStorableString("colorTweenInfoText", "<b>Color tweening</b> will change the emissive color between a set of 3 color values as set below.\nEnabling this will automatically disable 'Use Original Color' in the emissive plugin.");
                CreateTextField(pulseInfoJson, false);

                // 3 color pickers for tweening color
                HSVColor c = new HSVColor();
                c.H = 0.8f;
                c.S = 0.75f;
                c.V = 1.0f;
                _colorTweenA = new JSONStorableColor("TweenColor1", c);
                RegisterColor(_colorTweenA);
                var picker = CreateColorPicker(_colorTweenA, false);
                picker.label = "Tween Color 1";

                c = new HSVColor();
                c.H = 0.5f;
                c.S = 1.0f;
                c.V = 1.0f;
                _colorTweenB = new JSONStorableColor("TweenColor2", c);
                RegisterColor(_colorTweenB);
                picker = CreateColorPicker(_colorTweenB, false);
                picker.label = "Tween Color 2";

                c = new HSVColor();
                c.H = 0.15f;
                c.S = 0.5f;
                c.V = 1.0f;
                _colorTweenC = new JSONStorableColor("TweenColor3", c);
                RegisterColor(_colorTweenC);
                picker = CreateColorPicker(_colorTweenC, false);
                picker.label = "Tween Color 3";

                List<string> tweenTypes = new List<string>();
                tweenTypes.Add("Fixed");
                tweenTypes.Add("Lerp");
                tweenTypes.Add("Smooth");
                _colorTweenType = new JSONStorableStringChooser("ColorTweenType", tweenTypes, "Smooth", "colorTweenType", (string type) => { SettingsUpdated(); });
                RegisterStringChooser(_colorTweenType);
                var popup = CreateScrollablePopup(_colorTweenType, false);
                popup.label = "Color Tween Type";

                _colorTweenSpeed = new JSONStorableFloat("ColorTweenSpeedMs", 300.0f, (JSONStorableFloat jf) => { SettingsUpdated(); }, 10f, 1000f, false);
                RegisterFloat(_colorTweenSpeed);
                var slider = CreateSlider(_colorTweenSpeed, false);
                slider.valueFormat = "F1";
                slider.label = "Color Tween Speed (ms to next color)";
                slider.ConfigureQuickButtons(-1f, -10f, -50f, -100f, 1f, 10f, 50f, 100f);

                _colorTweenWrapHue = new JSONStorableBool("ColorTweenWrapHue", true, (bool value) => { SettingsUpdated(); });
                RegisterBool(_colorTweenWrapHue);
                toggle = CreateToggle(_colorTweenWrapHue, false);
                toggle.height = 80;
                toggle.label = "Wrap Hue";

                // preview color picker (readonly) for current color.
                c = new HSVColor();
                c.H = 0f;
                c.S = 1f;
                c.V = 1f;
                _previewCurrentTweenColor = new JSONStorableColor("PreviewTweenColor", c);
                picker = CreateColorPicker(_previewCurrentTweenColor, false);
                picker.label = "Color Tween (Preview)";
            }

            BuildPulseGUI();
            StartCoroutine(LoadAndInit());
        }

        protected IEnumerator LoadAndInit()
        {
            yield return WaitLoading();
            yield return new WaitForEndOfFrame();

            _emissivePlusScript = FindEmissiveClothingPlus();
            if (_emissivePlusScript == null)
            {
                SuperController.LogError("EmissiveExtraEffects: ");
            }

            _restoreFromJSONDone = true;
            _initScriptDone = true;
        }

        protected IEnumerator WaitLoading()
        {
            yield return new WaitForEndOfFrame();

            while (IsLoading())
            {
                yield return new WaitForEndOfFrame();
            }
        }

        protected bool IsLoading()
        {
            if (SuperController.singleton.isLoading)
                return true;

            Transform loadingIcon = SuperController.singleton.loadingIcon;
            if (loadingIcon == null || loadingIcon.gameObject == null)
                return true;

            if (loadingIcon.gameObject.active)
                return true;

            return false;
        }

        private bool _restoreFromJSONDone = false;
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            // Ensure restore from JSON is only done once and not by misstake later when load-merge another plugin.
            // If the load-merge plugin matches this one and this has not yet run its RestoreFromJSON, the load-merge
            // will call RestoreFromJSON on 2 plugin instances. This one and the newly added from the load-merge.
            if (!_restoreFromJSONDone)
            {
                _restoreFromJSONDone = true;
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
                BuildPulseGUI();
            }
        }

        void SettingsUpdated()
        {
            if (!_initScriptDone)
                return;

            _pulseCurrentStrength.min = _pulseMinValue.val;
            _pulseCurrentStrength.max = _pulseMaxValue.val;

            // Store / Restore original alpha value 
            // not saved, but makes it easier to "play around" with the effects
            if (_prevPulseActive != _pulseEnabled.val)
            {
                _prevPulseActive = _pulseEnabled.val;
                if (_emissivePlusScript != null)
                { 
                    if (_pulseEnabled.val)
                    {
                        _originalColorAlphaMultiplier = _emissivePlusScript.GetFloatParamValue("ColorAlphaMultiplier");
                    }
                    else if (_originalColorAlphaMultiplier.HasValue)
                    {
                        _emissivePlusScript.SetFloatParamValue("ColorAlphaMultiplier", _originalColorAlphaMultiplier.Value);
                    }
                }
            }

            // Store / Restore original color value
            // not saved, but makes it easier to "play around" with the effects
            if (_prevColorTweeningActive != _colorTweeningEnabled.val)
            {
                _prevColorTweeningActive = _colorTweeningEnabled.val;
                if (_emissivePlusScript != null)
                {
                    if (_colorTweeningEnabled.val)
                    {
                        _originalUseOriginalColor = _emissivePlusScript.GetBoolParamValue("UseOriginalColor");
                        _originalColor = _emissivePlusScript.GetColorParamValue("Color");
                    } 
                    else if (_originalUseOriginalColor.HasValue)
                    {
                        _emissivePlusScript.SetBoolParamValue("UseOriginalColor", _originalUseOriginalColor.Value);
                        _emissivePlusScript.SetColorParamValue("Color", _originalColor.Value);
                    }
                }
            }

            if (_colorTweeningEnabled.val && _emissivePlusScript != null)
            {
                _emissivePlusScript.SetBoolParamValue("UseOriginalColor", false);
            }
        }

        void BuildPulseGUI()
        {
            // remove old pulse GUI
            _pulseGUI.Destroy();

            // build new pulse GUI based on type
            switch (_pulseMainType.val)
            {
                case "Cyclic":
                    {
                        var slider = _pulseGUI.CreateSlider(_pulseMinValue, true);
                        slider.label = "Pulse Min Value";
                        slider.valueFormat = "F3";

                        slider = _pulseGUI.CreateSlider(_pulseMaxValue, true);
                        slider.label = "Pulse Max Value";
                        slider.valueFormat = "F3";

                        slider = _pulseGUI.CreateSlider(_pulseSpeed, true);
                        slider.label = "Pulse Speed (Pulses per second)";

                        slider = _pulseGUI.CreateSlider(_pulseCurrentStrength, true);
                        slider.label = "Pulse (Preview)";
                        slider.defaultButtonEnabled = false;
                        slider.quickButtonsEnabled = false;

                        _pulseInfoJson.val = "<b>Pulse</b> will update the emissive color alpha multiplier in the emissive script using a sinus curve. Adjust the min, max and speed properties below.";
                    }
                    break;
                case "ByMorphValue":
                    {
                        var popup = _pulseGUI.CreateFilterablePopup(_pulseMorphUidChooser, true);
                        popup.popupPanelHeight = 700f;
                        _pulseMorphUidChooser.popup.onOpenPopupHandlers += () => {
                            _pulseMorphUidChooser.choices = RebuildActiveMorphNames();
                        };

                        var toggle = _pulseGUI.CreateToggle(_pulseMorphInverse, true);
                        toggle.label = "Inverse Morph Value (1-morphValue)";
                        toggle.height = 80;

                        var slider = _pulseGUI.CreateSlider(_pulseMorphValueModifier, true);
                        slider.label = "Morph Value Modifier";

                        slider = _pulseGUI.CreateSlider(_pulseMorphValueOffset, true);
                        slider.label = "Morph Value Offset";

                        slider = _pulseGUI.CreateSlider(_pulseCurrentStrength, true);
                        slider.label = "Pulse (Preview)";
                        slider.defaultButtonEnabled = false;
                        slider.quickButtonsEnabled = false;

                        _pulseInfoJson.val = "Use a <b>morph value</b> to update the emissive color alpha multiplier in the emissive script. Only morphs with an active morph value is shown in the selector.";
                    }
                    break;
                case "ByFloatParamSource":
                    {
                        var popup = _pulseGUI.CreateFilterablePopup(_pulseAtomUidChooser, true);
                        popup.popupPanelHeight = 700f;
                        _pulseAtomUidChooser.popup.onOpenPopupHandlers += () => {
                            _pulseAtomUidChooser.choices = GetAtomChocies();
                        };

                        popup = _pulseGUI.CreateFilterablePopup(_pulseAtomStoreableUidChooser, true);
                        popup.popupPanelHeight = 700f;
                        _pulseAtomStoreableUidChooser.popup.onOpenPopupHandlers += () => {
                            _pulseAtomStoreableUidChooser.choices = GetAtomStoreableChoices(_pulseAtomUid.val);
                        };

                        popup = _pulseGUI.CreateFilterablePopup(_pulseAtomStoreableParamValueUidChooser, true);
                        popup.popupPanelHeight = 700f;
                        _pulseAtomStoreableParamValueUidChooser.popup.onOpenPopupHandlers += () => {
                            _pulseAtomStoreableParamValueUidChooser.choices = GetAtomStoreableFloatParamNames(_pulseAtomUid.val, _pulseAtomStoreableUid.val);
                        };

                        var toggle = _pulseGUI.CreateToggle(_pulseAtomStoreableParamInverse, true);
                        toggle.label = "Param Inverse Value (1-value)";
                        toggle.height = 80;

                        var slider = _pulseGUI.CreateSlider(_pulseAtomStoreableParamValueModifier, true);
                        slider.label = "Param Value Modifier";

                        slider = _pulseGUI.CreateSlider(_pulseAtomStoreableParamValueOffset, true);
                        slider.label = "Param Value Offset";

                        slider = _pulseGUI.CreateSlider(_pulseCurrentStrength, true);
                        slider.label = "Pulse (Preview)";
                        slider.defaultButtonEnabled = false;
                        slider.quickButtonsEnabled = false;

                        _pulseInfoJson.val = "Use a <b>float param value</b> from another atom storable/script to update the emissive color alpha multiplier in the emissive script.";
                    }
                    break;
            }
        }

        List<string> RebuildActiveMorphNames()
        {
            _filterMorphNames.Clear();
            _filterMorphUids.Clear();
            List<DAZMorph> morphs = new List<DAZMorph>();
            if (containingAtom != null)
            {
                DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
                if (geometry != null)
                {
                    foreach (DAZMorph morph in geometry.morphBank1.morphs)
                    {
                        if (morph != null && (morph.startValue != morph.appliedValue || morph.startValue != morph.morphValue))
                        {
                            morphs.Add(morph);
                        }
                    }
                    foreach (DAZMorph morph in geometry.morphBank2.morphs)
                    {
                        if (morph != null && (morph.startValue != morph.appliedValue || morph.startValue != morph.morphValue))
                        {
                            morphs.Add(morph);
                        }
                    }
                }
            }

            morphs.Sort((a, b) => { return string.Compare(a.displayName, b.displayName); });

            _filterMorphNames.Add("");
            _filterMorphUids.Add("");

            foreach (DAZMorph morph in morphs)
            {
                int dup = 0;
                string name = morph.displayName;
                while (_filterMorphNames.Contains(name))
                {
                    dup++;
                    name = $"{morph.displayName} ({dup.ToString()})";
                }
                _filterMorphNames.Add(name); // visible selector
                _filterMorphUids.Add(morph.uid); // matching uid
            }
            
            return _filterMorphNames;
        }

        void Update()
        {
            if (!_initScriptDone)
                return;

            UpdatePulse();
            UpdateColorTween();
        }

        void UpdatePulse()
        {
            float min = _pulseMinValue.val;
            float max = _pulseMaxValue.val;

            float targetStrength = 1f;

            switch (_pulseMainType.val)
            {
                case "Cyclic":
                    {
                        // controlled by sinus curve
                        _pulseTime += Time.deltaTime * PI2 * _pulseSpeed.val;
                        if (_pulseTime > PI2)
                            _pulseTime -= PI2;

                        float delta = (float)Math.Abs(max - min);
                        float strength = (float)(Math.Sin((double)_pulseTime) + 1.0) * 0.5f;
                        targetStrength = min + strength * delta;
                    }
                    break;
                case "ByMorphValue":
                    {
                        // controlled by morph
                        DAZMorph m = GetControlledByMorph();
                        if (m != null)
                        {
                            targetStrength = m.morphValue;
                        }

                        if (_pulseMorphInverse.val)
                        {
                            targetStrength = 1f - targetStrength;
                        }

                        targetStrength *= _pulseMorphValueModifier.val;

                        targetStrength += _pulseMorphValueOffset.val;
                    }
                    break;
                case "ByFloatParamSource":
                    {
                        // controlled by a float param source
                        JSONStorableFloat paramValue = GetSourceParamValue();
                        if (paramValue != null)
                        {
                            targetStrength = paramValue.val;
                        }

                        if (_pulseAtomStoreableParamInverse.val)
                        {
                            targetStrength = 1f - targetStrength;
                        }

                        targetStrength *= _pulseAtomStoreableParamValueModifier.val;

                        targetStrength += _pulseAtomStoreableParamValueOffset.val;
                    }
                    break;
            }

            if (_pulseCurrentStrength.min > targetStrength)
                _pulseCurrentStrength.min = targetStrength;
            if (_pulseCurrentStrength.max < targetStrength)
                _pulseCurrentStrength.max = targetStrength;

            _pulseCurrentStrength.val = targetStrength;

            if (_pulseEnabled.val && _emissivePlusScript != null)
            {
                _emissivePlusScript.SetFloatParamValue("ColorAlphaMultiplier", targetStrength);
            }
        }


        DAZMorph GetControlledByMorph()
        {
            string morphUid = _pulseMorphUid.val;
            if (string.IsNullOrEmpty(morphUid))
                return null;

            // If cached morph is null, only try to check for it each 1 seconds to reduce CPU usage
            // If a morph dependency package is not installed this will happen - this will likely be a common scenario for some
            if (_cachedPulseMorph == null && (Time.fixedTime - _cachedPulseMorphTime) < 1f)
                return null;

            // a valid cached morph exist (that's max 10 seconds old).
            if (_cachedPulseMorph != null && _cachedPulseMorph.uid == morphUid && (Time.fixedTime - _cachedPulseMorphTime) < 10f)
                return _cachedPulseMorph;

            // recache time
            _cachedPulseMorph = null;
            _cachedPulseMorphTime = Time.fixedTime;
            if (containingAtom != null)
            {
                DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
                if (geometry != null)
                {
                    foreach (DAZMorph morph in geometry.morphBank1.morphs)
                    {
                        if (morph != null && morph.uid == morphUid)
                        {
                            // SuperController.LogMessage("Cached morph = " + morph.uid);
                            _cachedPulseMorph = morph;
                            return morph;
                        }
                    }
                    foreach (DAZMorph morph in geometry.morphBank2.morphs)
                    {
                        if (morph != null && morph.uid == morphUid)
                        {
                            // SuperController.LogMessage("Cached morph = " + morph.uid);
                            _cachedPulseMorph = morph;
                            return morph;
                        }
                    }
                }
            }
            return null;
        }

        void UpdateColorTween()
        {
            
            HSVColor prevC = new HSVColor();
            HSVColor targetC = new HSVColor();
            float tweenSpeed = Math.Max(10f, _colorTweenSpeed.val);
            float totTween = tweenSpeed * 3f;
            float p = 1f;
            _tweenCurrentTime += Time.deltaTime * 1000f;
            if (_tweenCurrentTime >= totTween)
            {
                _tweenCurrentTime -= (int)(_tweenCurrentTime / totTween) * totTween;
            }

            if (_tweenCurrentTime <= tweenSpeed)
            {
                prevC = _colorTweenC.val;
                targetC = _colorTweenA.val;
                p = _tweenCurrentTime / tweenSpeed;
            }
            else if (_tweenCurrentTime <= tweenSpeed * 2f)
            {
                prevC = _colorTweenA.val;
                targetC = _colorTweenB.val;
                p = (_tweenCurrentTime - tweenSpeed) / tweenSpeed;
            }
            else
            {
                // From third to first
                prevC = _colorTweenB.val;
                targetC = _colorTweenC.val;
                p = (_tweenCurrentTime - tweenSpeed * 2f) / tweenSpeed;
            }

            switch (_colorTweenType.val)
            {
                case "Fixed": break;
                case "Lerp":
                    {
                        float dist = Math.Abs(targetC.H - prevC.H);
                        float wrapDist = 1f - dist;

                        if (dist < wrapDist || !_colorTweenWrapHue.val)
                        {
                            targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                        }
                        else
                        {
                            // increase and wrap-around or decrease and wrap-around?
                            if (prevC.H < targetC.H)
                            {
                                // decrease
                                float h = Mathf.Lerp(prevC.H, targetC.H - 1f, p);
                                if (h < 0f)
                                    h += 1f;

                                targetC.H = h;
                            }
                            else
                            {
                                // increase
                                float h = Mathf.Lerp(prevC.H, targetC.H + 1f, p);
                                if (h > 1f)
                                    h -= 1f;
                                targetC.H = h;
                            }
                        }

                        // targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                        targetC.S = Mathf.Lerp(prevC.S, targetC.S, p);
                        targetC.V = Mathf.Lerp(prevC.V, targetC.V, p);
                    }
                    break;
                case "Smooth":
                    {
                        float dist = Math.Abs(targetC.H - prevC.H);
                        float wrapDist = 1f - dist;

                        if (dist < wrapDist || !_colorTweenWrapHue.val)
                        {
                            targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                        }
                        else
                        {
                            // increase and wrap-around or decrease and wrap-around?
                            if (prevC.H < targetC.H)
                            {
                                // decrease
                                float h = Mathf.SmoothStep(prevC.H, targetC.H - 1f, p);
                                if (h < 0f)
                                    h += 1f;

                                targetC.H = h;
                            }
                            else
                            {
                                // increase
                                float h = Mathf.SmoothStep(prevC.H, targetC.H + 1f, p);
                                if (h > 1f)
                                    h -= 1f;
                                targetC.H = h;
                            }
                        }

                        // targetC.H = Mathf.SmoothStep(prevC.H, targetC.H, p);
                        targetC.S = Mathf.SmoothStep(prevC.S, targetC.S, p);
                        targetC.V = Mathf.SmoothStep(prevC.V, targetC.V, p);
                    }
                    break;
            }

            _previewCurrentTweenColor.val = targetC;

            if (_colorTweeningEnabled.val && _emissivePlusScript != null)
            {
                _emissivePlusScript.SetColorParamValue("Color", targetC);
            }
            
        }

        private EmissiveClothingPlus FindEmissiveClothingPlus()
        {
            string masterId = storeId.Substring(0, storeId.IndexOf("_") + 1) + typeof(EmissiveClothingPlus).FullName;
            return this.containingAtom.GetStorableByID(masterId) as EmissiveClothingPlus;
        }

        JSONStorableFloat GetSourceParamValue()
        {
            // cache matches?
            if (_sourceParamAtom != null && _sourceParamAtom.destroyed)
            {
                _sourceParamAtom = null;
                _sourceParamValue = null;
                _pulseParam_atomID = "";
                _pulseParam_storeableID = "";
                _pulseParam_paramID = "";
            }

            if (_pulseParam_atomID != _pulseAtomUid.val  ||
                _pulseParam_storeableID != _pulseAtomStoreableUid.val  ||
                _pulseParam_paramID != _pulseAtomStoreableParamValueUid.val)
            {
                _sourceParamValue = null;
                _pulseParam_atomID = _pulseAtomUid.val;
                _pulseParam_storeableID = _pulseAtomStoreableUid.val;
                _pulseParam_paramID = _pulseAtomStoreableParamValueUid.val;

                _sourceParamAtom = SuperController.singleton.GetAtomByUid(_pulseParam_atomID);
                if (_sourceParamAtom != null)
                {
                    JSONStorable storable = _sourceParamAtom.GetStorableByID(_pulseParam_storeableID);
                    if (storable != null)
                    {
                        _sourceParamValue = storable.GetFloatJSONParam(_pulseParam_paramID);
                    }

                }
            }

            return _sourceParamValue;
        }

        protected List<string> GetAtomChocies()
        {
            List<string> choices = new List<string>();
            choices.Add("");
            foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
            {
                choices.Add(atomUID);
            }
            return choices;
        }

        protected List<string> GetAtomStoreableChoices(string atomUID)
        {
            List<string> choices = new List<string>();
            choices.Add("");
            if (atomUID != null)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(atomUID);
                if (atom != null)
                {
                    foreach (string storableID in atom.GetStorableIDs())
                    {
                        choices.Add(storableID);
                        //SuperController.LogMessage("Found receiver " + receiverChoice);
                    }
                }
            }
            return choices;
        }

        protected List<string> GetAtomStoreableFloatParamNames(string atomID, string storeableID)
        {
            List<string> choices = new List<string>();
            choices.Add("");
            if (!string.IsNullOrEmpty(atomID) && !string.IsNullOrEmpty(storeableID))
            {
                Atom atom = SuperController.singleton.GetAtomByUid(atomID);
                if (atom != null)
                {
                    JSONStorable storable = atom.GetStorableByID(storeableID);
                    if (storable != null)
                    {
                        foreach (string floatParam in storable.GetFloatParamNames())
                        {
                            choices.Add(floatParam);
                        }
                    }
                }
            }
            return choices;
        }

    }
}